# 点
from .Point import Point
from .Points import Points

# 线
from .Polyline import Polyline
from .Circle import Circle

# 面
from .RectFace import RectFace

from .RPolygonFace import RPolygonFace
from .HollowRPolygonFace import HollowRPolygonFace

from .CircleFace import CircleFace
from .TorusFace import TorusFace
from .EllipseFace import EllipseFace

# 体

# 多面体
from .Box import Box

from .Prism import Prism
from .PrismSide import PrismSide

from .Pyramid import Pyramid
from .PyramidSide import PyramidSide

from .PyramidTrustum import PyramidTrustum
from .PyramidTrustumSide import PyramidTrustumSide

# 曲面体
from .Cylinder import Cylinder
from .CylinderSide import CylinderSide

from .Cone import Cone
from .ConeSide import ConeSide
from .ConeTrustum import ConeTrustum
from .ConeTrustumSide import ConeTrustumSide

from .Sphere import Sphere
from .Icosphere import Icosphere
from .SphericalCap import SphericalCap
from .SphericalCapTop import SphericalCapTop

# 正多面体
from .Tetrahedron import Tetrahedron # 正四面体
from .Hexahedron import Hexahedron # 正六面体
from .Octahedron import Octahedron # 正八面体
from .Dodecahedron import Dodecahedron # 正十二面体
from .Icosahedron import Icosahedron # 正二十面体

# 通用函数曲面
from .Surf import Surf
from .FSurf import FSurf
from .CylindricalFSurf import CylindricalFSurf
from .SphericalFSurf import SphericalFSurf

# 变形体
from .Rotator import Rotator
from .Extruder import Extruder

# 特殊几何体
from .Floor import Floor
from .Torus import Torus
from .TrefoilKnot import TrefoilKnot
from .CoordSys import CoordSys
from .ImageQuad import ImageQuad